#ifndef SimTK_DYNAMIC_SYSTEM_H_
#define SimTK_DYNAMIC_SYSTEM_H_

/* Portions copyright (c) 2007 Stanford University and Michael Sherman.
 * Contributors:
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including 
 * without limitation the rights to use, copy, modify, merge, publish, 
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject
 * to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS, COPYRIGHT HOLDERS, OR CONTRIBUTORS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/** @file
 * This header defines the SimTK Simmath notion of a "dynamic system" which
 * can be advanced through time by a SimTK::TimeStepper. The special case
 * of a fully continuous dynamic system is defined here also; those can
 * be advanced using only a SimTK::Integrator.
 */

#include "SimTKcommon.h"

#include <cmath>
#include <cassert>

/* Shared libraries are messy in Visual Studio. We have to distinguish three
 * cases:
 *   (1) this header is being used to build the SimTKcpodes shared library (dllexport)
 *   (2) this header is being used by a *client* of the SimTKcpodes shared
 *       library (dllimport)
 *   (3) we are building the SimTKcpodes static library, or the client is
 *       being compiled with the expectation of linking with the
 *       SimTKcpodes static library (nothing special needed)
 * In the CMake script for building this library, we define one of the symbols
 *     SimTK_CPODES_BUILDING_{SHARED|STATIC}_LIBRARY
 * Client code normally has no special symbol defined, in which case we'll
 * assume it wants to use the shared library. However, if the client defines
 * the symbol SimTK_USE_STATIC_LIBRARIES we'll suppress the dllimport so
 * that the client code can be linked with static libraries. Note that
 * the client symbol is not library dependent, while the library symbols
 * affect only the SimTKcpodes library, meaning that other libraries can
 * be clients of this one. However, we are assuming all-static or all-shared.
*/

// TODO: should be SIMMATH export
#ifdef WIN32
    #if defined(SimTK_CPODES_BUILDING_SHARED_LIBRARY)
        #define SimTK_CPODES_EXPORT __declspec(dllexport)
    #elif defined(SimTK_CPODES_BUILDING_STATIC_LIBRARY) || defined(SimTK_USE_STATIC_LIBRARIES)
        #define SimTK_CPODES_EXPORT
    #else
        /* i.e., a client of a shared library */
        #define SimTK_CPODES_EXPORT __declspec(dllimport)
    #endif
#else
    /* Linux, Mac */
    #define SimTK_CPODES_EXPORT
#endif

namespace SimTK {  


// TODO: move to SimTKcommon

// Sign transitions. Can be OR'ed together to make a SignTransitionSet.
class SignTransition {
public:
    enum SignTransitionBit {
        ZeroToPositive          =0x0001,    //  1
        ZeroToNegative          =0x0002,    //  2
        PositiveToZero          =0x0004,    //  4
        NegativeToZero          =0x0008,    //  8
        PositiveToNegative      =0x0010,    // 16
        NegativeToPositive      =0x0020,    // 32
    };

    // no default constructor

    // implicit conversion from SignTransitionBit
    SignTransition(SignTransitionBit bit) : transition(bit) { }
    SignTransition& operator=(SignTransitionBit bit) {
        transition=bit; return *this;
    }

    // implicit conversion to unsigned
    operator unsigned() const {return (unsigned)transition;}
private:
    SignTransitionBit transition;
};

class SignTransitionSet {
public:
    SignTransitionSet() : transitions(EmptySignTransitionSet) { }

    // implicit conversion from SignTransition to singleton set
    SignTransitionSet(SignTransition e) : transitions(e) { }

    // explicit conversion from unsigned
    explicit SignTransitionSet(unsigned s) : transitions(s) {
        assert( (s & AnySignChange) == s ); // watch for spurious bits
    }

    // implicit conversion to unsigned
    operator unsigned() const {return transitions;}

    void clear() {transitions=EmptySignTransitionSet;}

    // bitwise not here means all the transition bits which are
    // not in this set.
    SignTransitionSet operator~() const {
        return SignTransitionSet(AnySignChange & ~transitions);
    }

    SignTransitionSet& operator|=(SignTransitionSet es) {
        transitions |= es.transitions; return *this;
    }
    SignTransitionSet& operator&=(SignTransitionSet es) {
        transitions &= es.transitions; return *this;
    }
    SignTransitionSet& operator^=(SignTransitionSet es) {
        transitions ^= es.transitions; return *this;
    }
    SignTransitionSet& operator-=(SignTransitionSet es) {
        transitions &= ~es.transitions;
    }

    static const unsigned EmptySignTransitionSet = 0x0;                   // 0
    static const unsigned ZeroToNonzero =  SignTransition::ZeroToPositive // 3
                                          |SignTransition::ZeroToNegative;
    static const unsigned NonzeroToZero =  SignTransition::NegativeToZero // 12
                                          |SignTransition::PositiveToZero;
    static const unsigned PositiveToNonpositive                           // 20 
                                        =  SignTransition::PositiveToZero
                                          |SignTransition::PositiveToNegative;
    static const unsigned NegativeToNonnegative                           // 40
                                        =  SignTransition::NegativeToZero
                                          |SignTransition::NegativeToPositive;
    static const unsigned Rising        =  NegativeToNonnegative          // 41
                                          |SignTransition::ZeroToPositive;
    static const unsigned Falling       =  PositiveToNonpositive          // 22
                                          |SignTransition::ZeroToNegative;
    static const unsigned AnySignChange =  Rising | Falling;              // 63

private:
    unsigned transitions; // OR-ed together values from SignTransitionBit
};

/**
 * This abstract class defines a DynamicSystem to be advanced through time by a
 * TimeStepper object.
 *
 * The user's DynamicSystem object is const once constructed. However, it is expected
 * to be able to initialize a SimTK State object and then deal with updated
 * State objects passed to it by the TimeStepper.
 */
class SimTK_CPODES_EXPORT DynamicSystem {
public:

    // These are the exceptions that can be thrown by this class.
    class UserFunctionThrewAnException;
    class UserFunctionReturnedNonzeroStatus;
    class CantProjectStateOntoConstraints;

    const State& getDefaultState() const;
    State&       updDefaultState();

    /// Realize the given State to the indicated Stage, advancing
    /// through any unrealized lower stages first. Only the state
    /// cache can be modified; the state variables must be unchanged.
    virtual int realize(const State&, Stage) const=0;

        ///////////////////////////
        // THE CONTINUOUS SYSTEM //
        ///////////////////////////

            // UNCONSTRAINED

    /// Requires realize(Stage::Instance).
    virtual Real calcTimescale(const State&) const;

    /// Requires realize(Stage::Position).
    virtual int  calcWeights(const State&, Vector& weights) const;

            // CONSTRAINED
    
    /// Requires realize(Stage::Instance). Note that these are returned
    /// as 1/tol so that they can be used in the same manner as weights.
    virtual int  calcTolerances(const State&, Vector& ootols) const;

    /// This optional solver projects the given State back on to the constraint
    /// manifold, by the shortest path possible in the weighted norm
    /// given by the supplied weights, satisfying the constraints by
    /// reducing the supplied tolerance norm to below consAccuracy.
    /// May also project out the constraint-normal component of the
    /// passed-in error estimate vector yerrest.
    /// This is part of the integration of the continuous DAE system and
    /// thus should never require a restart. The System author must ensure
    /// that only position and velocity continuous variables are updated
    /// by this call.
    /// On return the state will be realized to at least Stage::Velocity.
    virtual int project(State&, Real consAccuracy, const Vector& yweights,
                        const Vector& ootols, Vector& yerrest) const;


        /////////////////////////
        // THE DISCRETE SYSTEM //
        /////////////////////////

    class EventTriggerInfo;

    /// This determines whether this System wants to be notified whenever time
    /// advances irreversibly. If set true, time advancement is treated as an
    /// event. Otherwise, time advancement proceeds silently.
    void setHasTimeAdvancedEvents(bool); // default=false
    bool hasTimeAdvancedEvents() const;


    /// These are all the possible causes for events. (1) An event trigger
    /// function may have undergone a monitored sign transition, (2) we
    /// may have reached the time for a scheduled event, (3) the integrator
    /// may have completed an internal step (assuming we registered our
    /// interest in those), (4) or a termination condition may have been
    /// detected. In case several of these causes are detected in 
    /// a single step, they are sequentialized in the order shown,
    /// like this:
    ///    1. The occurrence of triggered events is reported and
    ///       the triggering state and a list of triggered events
    ///       are passed to the event handler for processing (meaning
    ///       the state, but not the time, is modified). [Note that
    ///       simultaneity *within* the set of triggered events may
    ///       also require special handling; we're not talking about
    ///       that here, just simultaneity of *causes*.]
    ///    2. Next, using the state resulting from step 1, the time is checked
    ///       to see if scheduled events have occurred. If so, a list of
    ///       those events is passed to the event handler for processing.
    ///    3. Next, if this system has requested time-advanced events,
    ///       the event handler is called with the state that resulted
    ///       from step 2 and the "time advanced" cause noted. No event
    ///       list is passed in that case. The state may be modified.
    ///    4. Last, if the final time has been reached or if any of
    ///       the event handlers asked for termination, we pass the
    ///       state to the event handler again noting that we have
    ///       reached termination. The state may be modified and the
    ///       result will be the final state of the simulation.

    enum EventCause {
        TriggeredEvents    =1,
        ScheduledEvents    =2,
        TimeAdvancedEvent  =3,
        TerminationEvent   =4,

        InvalidEventCause  = -1
    };
    static const char* getEventCauseName(EventCause);


    /// This solver handles a set of events which the time stepper has
    /// denoted as having occurred. The event handler may make discontinuous
    /// changes in the State, in general both to discrete and continuous variables,
    /// but NOT to time. If changes are made to continuous variables, the handler
    /// is required to make sure the returned state satisfies the constraints to the
    /// indicated accuracy level.
    /// On return, the handleEvents routine should set the output variable
    /// lowestModified to the Stage level of the lowest-stage variable
    /// it modified. This information tells the time stepper how much of 
    /// a restart it must perform on the underlying numerical integrator. 
    /// When in doubt, set lowestModified to Stage::Model, which will cause
    /// a complete restart.
    /// Finally, if the handler determines that the occurrence of some event
    /// requires that the simulation be terminated it should set 'shouldTerminate'
    /// to true before returning.

    virtual int handleEvents
       (State&, EventCause, const Array<int>& eventIds,
        Real accuracy, const Vector& yWeights, const Vector& ooConstraintTols,
        Stage& lowestModified, bool& shouldTerminate) const;

    /// This routine provides the Integrator with information it needs about the
    /// individual event trigger functions, such as which sign transitions are
    /// relevant and how tightly we need to localize. This is considered 
    /// Instance stage information so cannot change during a continuous integration
    /// interval (so an Integrator can process it upon restart(Instance)), however it can be
    /// updated whenever a discrete update is made to the State.
    /// A default implementation is provided which returns default EventTriggerInfo for
    /// each event trigger in State.
    /// State must already be realized to Stage::Instance.
    virtual int calcEventTriggerInfo(const State&, Array<EventTriggerInfo>&) const;

    /// This routine should be called to determine if and when there is an event
    /// scheduled to occur at a particular time. This is a *lot* cheaper than
    /// making the integrator hunt these down like ordinary state-dependent events.
    /// The returned time can be passed to the integrator's stepping function as
    /// the advance time limit.
    /// The default implementation of this virtual method return infinity as the
    /// next scheduled time.
    virtual int calcTimeOfNextScheduledEvent
        (const State&, Real& tNextEvent, Array<int>& eventIds) const;

    // TODO: these utilities don't really belong here
    static Real calcWeightedRMSNorm(const Vector& values, const Vector& weights) {
        assert(weights.size() == values.size());
        if (values.size()==0) return 0;
        Real sumsq = 0;
        for (int i=0; i<values.size(); ++i) {
            const Real wv = values[i]*weights[i];
            sumsq += wv*wv;
        }
        return std::sqrt(sumsq/weights.size());
    }

    static Real calcWeightedInfinityNorm(const Vector& values, const Vector& weights) {
        assert(weights.size() == values.size());
        if (values.size()==0) return 0;
        Real maxval = 0;
        for (int i=0; i<values.size(); ++i) {
            const Real wv = std::abs(values[i]*weights[i]);
            if (wv > maxval) maxval=wv;
        }
        return maxval;
    }

    // These typedefs are used internally to manage the binary-compatible
    // handling of the virtual function table.

    typedef int  (*RealizeWrapper)(const DynamicSystem&, const State&, Stage);
    typedef Real (*CalcTimescaleWrapper)(const DynamicSystem&, const State&);
    typedef int  (*CalcWeightsWrapper)(const DynamicSystem&, const State&, Vector&);

    typedef int  (*CalcTolerancesWrapper)(const DynamicSystem&, const State&, Vector&);
    typedef int  (*ProjectWrapper)(const DynamicSystem&, State&, Real, const Vector&, const Vector&,
                                   Vector&);

    typedef int  (*HandleEventsWrapper)
       (const DynamicSystem&, State&, EventCause, const Array<int>&,
        Real, const Vector&, const Vector&, Stage&, bool&);
    typedef int  (*CalcEventTriggerInfoWrapper)
       (const DynamicSystem&, const State&, Array<EventTriggerInfo>&);
    typedef int  (*CalcTimeOfNextScheduledEventWrapper)
       (const DynamicSystem&, const State&, Real&, Array<int>&);

    class EventTriggerInfoRep;

    // must be inline for binary compatibility
    inline explicit DynamicSystem(); 
    inline virtual ~DynamicSystem() {librarySideDestruction();}

    // opaque implementation for binary compatibility
    class DynamicSystemRep* rep;

    const DynamicSystemRep& getRep() const {assert(rep); return *rep;}
    DynamicSystemRep&       updRep()       {assert(rep); return *rep;}

private:
    void librarySideConstruction();
    void librarySideDestruction();

    void registerRealize(RealizeWrapper);
    void registerCalcTimescale(CalcTimescaleWrapper);
    void registerCalcWeights(CalcWeightsWrapper);

    void registerCalcTolerances(CalcTolerancesWrapper);
    void registerProject(ProjectWrapper);

    void registerHandleEvents(HandleEventsWrapper);
    void registerCalcEventTriggerInfo(CalcEventTriggerInfoWrapper);
    void registerCalcTimeOfNextScheduledEvent(CalcTimeOfNextScheduledEventWrapper);

    // suppress copy constructor and copy assignment
    DynamicSystem(const DynamicSystem&);
    DynamicSystem& operator=(const DynamicSystem&);
};


// These are used to supply the client-side virtual function to the library, without
// the client and library having to agree on the layout of the virtual function tables.

static int dynamicSystemRealizeWrapper
   (const DynamicSystem& sys, const State& state, Stage g)
{ return sys.realize(state,g); }

static Real dynamicSystemCalcTimescaleWrapper
   (const DynamicSystem& sys, const State& state)
{ return sys.calcTimescale(state); }

static int dynamicSystemCalcWeightsWrapper
   (const DynamicSystem& sys, const State& state, Vector& weights)
{ return sys.calcWeights(state,weights); }

static int dynamicSystemCalcTolerancesWrapper
   (const DynamicSystem& sys, const State& state, Vector& tols)
{ return sys.calcTolerances(state,tols); }

static int dynamicSystemProjectWrapper
   (const DynamicSystem& sys, State& state, Real acc, const Vector& wt, const Vector& tol,
    Vector& yerrest)
{ return sys.project(state,acc,wt,tol,yerrest); }

static int dynamicSystemHandleEventsWrapper
   (const DynamicSystem& sys, State& state, DynamicSystem::EventCause cause, const Array<int>& eventIds,
    Real accuracy, const Vector& yWeights, const Vector& ooConstraintTols,
    Stage& lowestModified, bool& shouldTerminate)
{ 
    return sys.handleEvents(state,cause,eventIds,accuracy,yWeights,ooConstraintTols,
                            lowestModified, shouldTerminate);
}

static int dynamicSystemCalcEventTriggerInfoWrapper
   (const DynamicSystem& sys, const State& state, Array<DynamicSystem::EventTriggerInfo>& info)
{ 
    return sys.calcEventTriggerInfo(state,info);
}

static int dynamicSystemCalcTimeOfNextScheduledEventWrapper
   (const DynamicSystem& sys, const State& state, Real& tNextEvent, Array<int>& eventIds)
{ 
    return sys.calcTimeOfNextScheduledEvent(state,tNextEvent,eventIds);
}

inline DynamicSystem::DynamicSystem() : rep(0)
{
    librarySideConstruction();

    registerRealize(dynamicSystemRealizeWrapper);
    registerCalcTimescale(dynamicSystemCalcTimescaleWrapper);
    registerCalcWeights(dynamicSystemCalcWeightsWrapper);

    registerCalcTolerances(dynamicSystemCalcTolerancesWrapper);
    registerProject(dynamicSystemProjectWrapper);

    registerHandleEvents(dynamicSystemHandleEventsWrapper);
    registerCalcEventTriggerInfo
       (dynamicSystemCalcEventTriggerInfoWrapper);
    registerCalcTimeOfNextScheduledEvent
       (dynamicSystemCalcTimeOfNextScheduledEventWrapper);
}

/// This class is used to communicate between the DynamicSystem and the 
/// Integrator regarding the properties of a particular event trigger
/// function. Currently these are:
///   - Whether to watch for rising sign transitions, falling, or both. [BOTH]
///   - Whether to watch for transitions to and from zero. [NO]
///   - The localization window in units of the System's timescale. [1%]
/// The default values are shown in brackets above.
///
class SimTK_CPODES_EXPORT DynamicSystem::EventTriggerInfo {
public:
    EventTriggerInfo();
    explicit EventTriggerInfo(int eventId);
    ~EventTriggerInfo();
    EventTriggerInfo(const EventTriggerInfo&);
    EventTriggerInfo& operator=(const EventTriggerInfo&);

    int getEventId() const; // returns -1 if not set
    bool shouldTriggerOnRisingSignTransition()  const; // default=true
    bool shouldTriggerOnFallingSignTransition() const; // default=true
    bool shouldTriggerOnZeroTransitions()       const; // default=false
    Real getRequiredLocalizationTimeWindow()    const; // default=0.01

    // These return the modified 'this', like assignment operators.
    EventTriggerInfo& setEventId(int);
    EventTriggerInfo& setTriggerOnRisingSignTransition(bool);
    EventTriggerInfo& setTriggerOnFallingSignTransition(bool);
    EventTriggerInfo& setTriggerOnZeroTransitions(bool);
    EventTriggerInfo& setRequiredLocalizationTimeWindow(Real);

private:
    // opaque implementation for future binary compatibility
    DynamicSystem::EventTriggerInfoRep* rep;

    const DynamicSystem::EventTriggerInfoRep& getRep() const {assert(rep); return *rep;}
    DynamicSystem::EventTriggerInfoRep&       updRep()       {assert(rep); return *rep;}
};

// Global operators on SignTransitionSets (there is an implicit conversion
// from SignTransition to singleton SignTransitionSet when needed).

/// Return the union of two SignTransitionSets.
inline SignTransitionSet operator|(SignTransitionSet sts1, SignTransitionSet sts2) {
    return sts1 |= sts2;
}

/// Return the intersection of two SignTransitionSets.
inline SignTransitionSet operator&(SignTransitionSet sts1, SignTransitionSet sts2) {
    return sts1 &= sts2;
}

/// Return the SignTransitionSet containing the SignTransitions that are in
/// one of two sets but not both (exclusive or).
inline SignTransitionSet operator^(SignTransitionSet sts1, SignTransitionSet sts2) {
    return sts1 ^= sts2;
}

/// Return the SignTransitionSet containing the SignTransitions that are in
/// the left hand set but not also in the right hand set (set difference).
inline SignTransitionSet operator-(SignTransitionSet sts1, SignTransitionSet sts2) {
    return sts1 -= sts2;
}


} // namespace SimTK

#endif // SimTK_DYNAMIC_SYSTEM_H_
