#ifndef SimTK_TIMESTEPPER_H_
#define SimTK_TIMESTEPPER_H_

/* Portions copyright (c) 2007 Stanford University and Michael Sherman.
 * Contributors:
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including 
 * without limitation the rights to use, copy, modify, merge, publish, 
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject
 * to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included 
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS, COPYRIGHT HOLDERS, OR CONTRIBUTORS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/** @file
 * This is the header file that user code should include to pick up the 
 * SimTK Simmath "time stepper" tools.
 */

#include "SimTKcommon.h"

/* Shared libraries are messy in Visual Studio. We have to distinguish three
 * cases:
 *   (1) this header is being used to build the SimTKcpodes shared library (dllexport)
 *   (2) this header is being used by a *client* of the SimTKcpodes shared
 *       library (dllimport)
 *   (3) we are building the SimTKcpodes static library, or the client is
 *       being compiled with the expectation of linking with the
 *       SimTKcpodes static library (nothing special needed)
 * In the CMake script for building this library, we define one of the symbols
 *     SimTK_CPODES_BUILDING_{SHARED|STATIC}_LIBRARY
 * Client code normally has no special symbol defined, in which case we'll
 * assume it wants to use the shared library. However, if the client defines
 * the symbol SimTK_USE_STATIC_LIBRARIES we'll suppress the dllimport so
 * that the client code can be linked with static libraries. Note that
 * the client symbol is not library dependent, while the library symbols
 * affect only the SimTKcpodes library, meaning that other libraries can
 * be clients of this one. However, we are assuming all-static or all-shared.
*/

// TODO: should be SIMMATH export
#ifdef WIN32
    #if defined(SimTK_CPODES_BUILDING_SHARED_LIBRARY)
        #define SimTK_CPODES_EXPORT __declspec(dllexport)
    #elif defined(SimTK_CPODES_BUILDING_STATIC_LIBRARY) || defined(SimTK_USE_STATIC_LIBRARIES)
        #define SimTK_CPODES_EXPORT
    #else
        /* i.e., a client of a shared library */
        #define SimTK_CPODES_EXPORT __declspec(dllimport)
    #endif
#else
    /* Linux, Mac */
    #define SimTK_CPODES_EXPORT
#endif

namespace SimTK {
class DynamicSystem;
class Integrator;

/**
 * Given a system of equations, an object of this class will advance that
 * system through time. The system is expected to be a mixture of continuous
 * and discrete equations, involving continuous state variables y and 
 * discrete state variables d. The continuous part is an ODE-on-a-manifold
 * system suitable for solution via coordinate projection[1], structured like
 * this:
 *         (1)  y' = f(d;t,y)         differential equations
 *         (2)  c  = c(d;t,y)         algebraic equations (manifold is c=0)
 *         (3)  e  = e(d;t,y)         event triggers (watch for zero crossings)
 * with initial conditions t0,y0,d0 such that c=0. By "ODE on a manifold" we
 * mean that the ODE (1) automatically satisfies the condition that IF c==0,
 * THEN c'=0, where c'=partial(c)/partial(t) + [partial(c)/partial(y)]*y'. This is
 * a less stringent condition than an ODE with invariant, in which c'=0 regardless
 * of whether c=0.
 *
 * [1] Hairer, Lubich, Wanner, "Geometric Numerical Integration: Structure-Preserving
 * Algorithms for Ordinary Differential Equations", 2nd ed., section IV.4, pg 109ff,
 * Springer, 2006.
 *
 * The discrete variables d are updated upon occurence of specific events, which are
 * detected using a set of scalar-valued event trigger functions (3).
 * An event trigger function for a particular event should be designed so that it has a
 * zero crossing when the event occurs. The integrator can thus watch for sign changes
 * in event triggers and terminate the current step when a zero crossing occurs,
 * notifying the system and giving it a chance to handle the event; that is, 
 * update its state variables discontinuously.
 *
 * The zero crossings of continuous event trigger functions will
 * be isolated quickly; discontinuous ones have to be "binary chopped" which
 * is more expensive. There are several special case events:
 *          - events which are simply a known function of t
 *          - "end of step" events
 *          - external events (e.g., clock time, user interrupt)
 *
 * We are given a set of weights W for the y's, and a set of tolerances T
 * for the constraint errors. Given an accuracy specification (like 0.1%),
 * the integrators here are expected to solve for y(t) such that the
 * local error |W*y|_RMS <= accuracy, and |T*c(t,y)|_RMS <= accuracy at
 * all times.
 *
 * TODO: isolation tolerances for witnesses; dealing with simultaneity.
 *
 */
class SimTK_CPODES_EXPORT TimeStepper {
public:
    explicit TimeStepper(const DynamicSystem&);
    ~TimeStepper();

    TimeStepper& setIntegrator(const Integrator&);
    const Integrator& getIntegrator() const;
    Integrator& updIntegrator();

    // Supply the time stepper with a starting state. This is *copied*
    // into the time stepper's internally maintained "advanced" state; 
    // subsequent changes to the State object passed in here will not
    // affect the simulation.
    void initialize(const State&);

    // Return a state which satisfies the caller's step request. This 
    // may be an interpolated value earlier than getAdvancedState().
    const State& getState() const;
    Real         getTime() const {return getState().getTime();}

    // Control of time stepping (behavior of stepTo() method).
    TimeStepper& setAccuracy(Real);       // easiest way to control cost vs. accuracy; default=1e-3
    TimeStepper& setFinalTime(Real);      // time won't advance past here; default=infinity
    TimeStepper& setReportAtEvents(bool); // should stepTo return before/after event processing? default=true


    // When a step is successful, it will return an indication of what
    // caused it to stop where it did. When unsuccessful it will throw
    // an exception so you won't see any return value.
    //
    // When return of control is due ONLY to reaching a report time,
    // (status is ReachedReportTime) the time stepper's getState() method may return an
    // interpolated value at an earlier time than its getAdvancedState()
    // method would return. For the other returns, getState() and
    // getAdvancedState() will be identical.
    //
    // Note that if some other termination condition happens to occur
    // exactly at a report time, the return status will be the other condition.

    enum SuccessfulStepStatus {
        IntervalStartState    =1, // start of simulation or post-event handling
        ReachedReportTime     =2, // stopped only to report; might be interpolated
        ReachedScheduledEvent =3, // reached a pre-scheduled time
        ReachedTriggeredEvent =4, // localized an event; this is the *before* state
        ReachedStepLimit      =5, // took a lot of internal steps but didn't reach reportTime
        ReachedFinalTime      =6, // hit user-specified max time; don't call again

        InvalidSuccessfulStepStatus = -1
    };

    SuccessfulStepStatus stepTo(Real reportTime);
    SuccessfulStepStatus stepBy(Real interval);

    static String successfulStepStatusString(SuccessfulStepStatus status);

    // Statistics (mutable)
    void resetAllStatistics();                // reset all stats to zero

    
    //class HybridSystemRep;
    class EventInfoRep;
private:
    // opaque implementation for binary compatibility
    class TimeStepperRep* rep;
    friend class TimeStepperRep;
};

} // namespace SimTK

#endif // SimTK_TIMESTEPPER_H_
